@@ -24,13 +24,12 @@ License along with TPOT. If not, see <http://www.gnu.org/licenses/>.
 """
 
 from tpot import TPOTClassifier, TPOTRegressor
-from tpot.base import TPOTBase, is_notebook
+from tpot.base import TPOTBase
 from tpot.driver import float_range
 from tpot.gp_types import Output_Array
 from tpot.gp_deap import mutNodeReplacement, _wrapped_cross_val_score, pick_two_individuals_eligible_for_crossover, cxOnePoint, varOr, initialize_stats_dict
 from tpot.metrics import balanced_accuracy, SCORERS
-from tpot.operator_utils import TPOTOperatorClassFactory, set_sample_weight
-from tpot.decorators import pretest_X, pretest_y
+from tpot.operator_utils import TPOTOperatorClassFactory, set_sample_weight, source_decode
 
 from tpot.config.classifier import classifier_config_dict
 from tpot.config.classifier_light import classifier_config_dict_light
@@ -54,16 +53,22 @@ from time import sleep
 from tempfile import mkdtemp
 from shutil import rmtree
 
-from sklearn.datasets import load_digits, load_boston
+from sklearn.datasets import load_digits, load_boston, make_classification
 from sklearn.model_selection import train_test_split, cross_val_score, GroupKFold
-from sklearn.externals.joblib import Memory
+from joblib import Memory
 from sklearn.metrics import make_scorer, roc_auc_score
+from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin, TransformerMixin
+from sklearn.feature_selection.base import SelectorMixin
 from deap import creator, gp
 from deap.tools import ParetoFront
-from nose.tools import assert_raises, assert_not_equal, assert_greater_equal, assert_equal, assert_in
+from nose.tools import nottest, assert_raises, assert_not_equal, assert_greater_equal, assert_equal, assert_in
 from driver_tests import captured_output
 
-from tqdm import tqdm
+train_test_split = nottest(train_test_split)
+
+with warnings.catch_warnings():
+    warnings.simplefilter('ignore')
+    from tqdm.autonotebook import tqdm
 
 try:
     from StringIO import StringIO
@@ -92,6 +97,9 @@ boston_data = load_boston()
 training_features_r, testing_features_r, training_target_r, testing_target_r = \
     train_test_split(boston_data.data, boston_data.target, random_state=42)
 
+# Set up a small test dataset
+
+pretest_X, pretest_y = make_classification(n_samples=100, n_features=10, random_state=42)
 # Set up pandas DataFrame for testing
 
 input_data = pd.read_csv(
@@ -154,13 +162,7 @@ def test_init_custom_parameters():
     assert tpot_obj._optimized_pipeline == None
     assert tpot_obj._optimized_pipeline_score == None
     assert tpot_obj.fitted_pipeline_ == None
-    assert tpot_obj._exported_pipeline_text == ""
-
-
-def test_is_notebook():
-    """Assert that isnotebook function works as expected."""
-    ret = is_notebook()
-    assert not ret
+    assert tpot_obj._exported_pipeline_text == []
 
 
 def test_init_default_scoring():
@@ -225,7 +227,7 @@ def test_init_default_scoring_6():
     assert len(w) == 1
     assert issubclass(w[-1].category, DeprecationWarning)
     assert "This scoring type was deprecated" in str(w[-1].message)
-    print(tpot_obj.scoring_function)
+
     assert tpot_obj.scoring_function == 'my_scorer'
 
 
@@ -283,6 +285,8 @@ def test_init_n_jobs():
     """Assert that the TPOT init stores current number of processes."""
     tpot_obj = TPOTClassifier(n_jobs=2)
     assert tpot_obj.n_jobs == 2
+    tpot_obj._fit_init()
+    assert tpot_obj._n_jobs == 2
 
     tpot_obj = TPOTClassifier(n_jobs=-1)
     assert tpot_obj.n_jobs == -1
@@ -290,6 +294,23 @@ def test_init_n_jobs():
     assert tpot_obj._n_jobs == cpu_count()
 
 
+def test_init_n_jobs_2():
+    """Assert that the TPOT init assign right"""
+    tpot_obj = TPOTClassifier(n_jobs=-2)
+    assert tpot_obj.n_jobs == -2
+
+    tpot_obj._fit_init()
+    assert tpot_obj._n_jobs == cpu_count() - 1
+
+
+def test_init_n_jobs_3():
+    """Assert that the TPOT init rasies ValueError if n_jobs=0."""
+    tpot_obj = TPOTClassifier(n_jobs=0)
+    assert tpot_obj.n_jobs == 0
+
+    assert_raises(ValueError, tpot_obj._fit_init)
+
+
 def test_timeout():
     """Assert that _wrapped_cross_val_score return Timeout in a time limit."""
     tpot_obj = TPOTRegressor(scoring='neg_mean_squared_error')
@@ -527,7 +548,7 @@ def test_score_3():
     """Assert that the TPOTRegressor score function outputs a known score for a fixed pipeline."""
     tpot_obj = TPOTRegressor(scoring='neg_mean_squared_error', random_state=72)
     tpot_obj._fit_init()
-    known_score = -12.1791953611
+    known_score = -11.682841148312662
 
     # Reify pipeline with known score
     pipeline_string = (
@@ -590,14 +611,99 @@ def test_sample_weight_func():
     np.random.seed(42)
     tpot_obj.fitted_pipeline_.fit(training_features_r, training_target_r, **training_target_r_weight_dict)
     # Get score from TPOT
-    known_score = -11.5790430757
+    known_score = -11.586816877933911
     score = tpot_obj.score(testing_features_r, testing_target_r)
 
+
     assert np.allclose(cv_score1, cv_score2)
     assert not np.allclose(cv_score1, cv_score_weight)
     assert np.allclose(known_score, score)
 
 
+def test_template_1():
+    """Assert that TPOT template option generates pipeline when each step is a type of operator."""
+
+    tpot_obj = TPOTClassifier(
+        random_state=42,
+        verbosity=0,
+        template='Selector-Transformer-Classifier'
+    )
+    tpot_obj._fit_init()
+    pop = tpot_obj._toolbox.population(n=10)
+    for deap_pipeline in pop:
+        operator_count = tpot_obj._operator_count(deap_pipeline)
+        sklearn_pipeline = tpot_obj._toolbox.compile(expr=deap_pipeline)
+        assert operator_count == 3
+        assert issubclass(sklearn_pipeline.steps[0][1].__class__, SelectorMixin)
+        assert issubclass(sklearn_pipeline.steps[1][1].__class__, TransformerMixin)
+        assert issubclass(sklearn_pipeline.steps[2][1].__class__, ClassifierMixin)
+
+
+def test_template_2():
+    """Assert that TPOT template option generates pipeline when each step is operator type with a duplicate main type."""
+
+    tpot_obj = TPOTClassifier(
+        random_state=42,
+        verbosity=0,
+        template='Selector-Selector-Transformer-Classifier'
+    )
+    tpot_obj._fit_init()
+    pop = tpot_obj._toolbox.population(n=10)
+    for deap_pipeline in pop:
+        operator_count = tpot_obj._operator_count(deap_pipeline)
+        sklearn_pipeline = tpot_obj._toolbox.compile(expr=deap_pipeline)
+        assert operator_count == 4
+        assert issubclass(sklearn_pipeline.steps[0][1].__class__, SelectorMixin)
+        assert issubclass(sklearn_pipeline.steps[1][1].__class__, SelectorMixin)
+        assert issubclass(sklearn_pipeline.steps[2][1].__class__, TransformerMixin)
+        assert issubclass(sklearn_pipeline.steps[3][1].__class__, ClassifierMixin)
+
+
+def test_template_3():
+    """Assert that TPOT template option generates pipeline when one of steps is a specific operator."""
+
+    tpot_obj = TPOTClassifier(
+        random_state=42,
+        verbosity=0,
+        template='SelectPercentile-Transformer-Classifier'
+    )
+    tpot_obj._fit_init()
+    pop = tpot_obj._toolbox.population(n=10)
+    for deap_pipeline in pop:
+        operator_count = tpot_obj._operator_count(deap_pipeline)
+        sklearn_pipeline = tpot_obj._toolbox.compile(expr=deap_pipeline)
+        assert operator_count == 3
+        assert sklearn_pipeline.steps[0][0] == 'SelectPercentile'.lower()
+        assert issubclass(sklearn_pipeline.steps[0][1].__class__, SelectorMixin)
+        assert issubclass(sklearn_pipeline.steps[1][1].__class__, TransformerMixin)
+        assert issubclass(sklearn_pipeline.steps[2][1].__class__, ClassifierMixin)
+
+
+def test_template_4():
+    """Assert that TPOT template option generates pipeline when one of steps is a specific operator."""
+
+    tpot_obj = TPOTClassifier(
+        population_size=5,
+        generations=2,
+        random_state=42,
+        verbosity=0,
+        config_dict = 'TPOT light',
+        template='SelectPercentile-Transformer-Classifier'
+    )
+    tpot_obj.fit(pretest_X, pretest_y)
+
+    assert isinstance(tpot_obj._optimized_pipeline, creator.Individual)
+    assert not (tpot_obj._start_datetime is None)
+
+    sklearn_pipeline = tpot_obj.fitted_pipeline_
+    operator_count = tpot_obj._operator_count(tpot_obj._optimized_pipeline)
+    assert operator_count == 3
+    assert sklearn_pipeline.steps[0][0] == 'SelectPercentile'.lower()
+    assert issubclass(sklearn_pipeline.steps[0][1].__class__, SelectorMixin)
+    assert issubclass(sklearn_pipeline.steps[1][1].__class__, TransformerMixin)
+    assert issubclass(sklearn_pipeline.steps[2][1].__class__, ClassifierMixin)
+
+
 def test_fit_GroupKFold():
     """Assert that TPOT properly handles the group parameter when using GroupKFold."""
     # This check tests if the darker MNIST images would generalize to the lighter ones.
@@ -936,18 +1042,23 @@ def test_memory_3():
 
 
 def test_memory_4():
-    """Assert that the TPOT _setup_memory function rasies ValueError with a invalid path."""
+    """Assert that the TPOT _setup_memory function create a directory which does not exist."""
+    cachedir = mkdtemp()
+    dir = cachedir + '/test'
     tpot_obj = TPOTClassifier(
         random_state=42,
         population_size=1,
         offspring_size=2,
         generations=1,
         config_dict='TPOT light',
-        memory="./fake_temp_dir",
+        memory=dir,
         verbosity=0
     )
+    tpot_obj._setup_memory()
+    assert os.path.isdir(dir)
+    rmtree(cachedir)
+
 
-    assert_raises(ValueError, tpot_obj._setup_memory)
 
 
 def test_memory_5():
@@ -1005,48 +1116,16 @@ def test_check_periodic_pipeline():
         tpot_obj._last_pipeline_write = datetime.now()
         sleep(0.11)
         tpot_obj._output_best_pipeline_period_seconds = 0.1
-        tpot_obj.periodic_checkpoint_folder = './'
-        tpot_obj._check_periodic_pipeline()
+        tmpdir = mkdtemp() + '/'
+        tpot_obj.periodic_checkpoint_folder = tmpdir
+        tpot_obj._check_periodic_pipeline(1)
         our_file.seek(0)
-
-        assert_in('Saving best periodic pipeline to ./pipeline', our_file.read())
+        assert_in('Saving periodic pipeline from pareto front', our_file.read())
         # clean up
-        for f in os.listdir('./'):
-            if search('pipeline_', f):
-                os.remove(os.path.join('./', f))
+        rmtree(tmpdir)
 
 
 def test_check_periodic_pipeline_2():
-    """Assert that the _check_periodic_pipeline does not export periodic pipeline if the pipeline has been saved before."""
-    tpot_obj = TPOTClassifier(
-        random_state=42,
-        population_size=1,
-        offspring_size=2,
-        generations=1,
-        verbosity=0,
-        config_dict='TPOT light'
-    )
-    tpot_obj.fit(training_features, training_target)
-    with closing(StringIO()) as our_file:
-        tpot_obj._file = our_file
-        tpot_obj.verbosity = 3
-        tpot_obj._last_pipeline_write = datetime.now()
-        sleep(0.11)
-        tpot_obj._output_best_pipeline_period_seconds = 0.1
-        tpot_obj.periodic_checkpoint_folder = './'
-        # export once before
-        tpot_obj.export('./pipeline_test.py')
-        tpot_obj._check_periodic_pipeline()
-        our_file.seek(0)
-
-        assert_in('Periodic pipeline was not saved, probably saved before...', our_file.read())
-        # clean up
-        for f in os.listdir('./'):
-            if search('pipeline_', f):
-                os.remove(os.path.join('./', f))
-
-
-def test_check_periodic_pipeline_3():
     """Assert that the _check_periodic_pipeline rasie StopIteration if self._last_optimized_pareto_front_n_gens >= self.early_stop."""
     tpot_obj = TPOTClassifier(
         random_state=42,
@@ -1059,9 +1138,9 @@ def test_check_periodic_pipeline_3():
     tpot_obj.fit(training_features, training_target)
     tpot_obj.early_stop = 3
     # will pass
-    tpot_obj._check_periodic_pipeline()
+    tpot_obj._check_periodic_pipeline(1)
     tpot_obj._last_optimized_pareto_front_n_gens = 3
-    assert_raises(StopIteration, tpot_obj._check_periodic_pipeline)
+    assert_raises(StopIteration, tpot_obj._check_periodic_pipeline, 1)
 
 
 def test_save_periodic_pipeline():
@@ -1081,18 +1160,18 @@ def test_save_periodic_pipeline():
         tpot_obj._last_pipeline_write = datetime.now()
         sleep(0.11)
         tpot_obj._output_best_pipeline_period_seconds = 0.1
-        tpot_obj.periodic_checkpoint_folder = './'
-        # reset _optimized_pipeline to rasie exception
-        tpot_obj._optimized_pipeline = None
+        tmpdir = mkdtemp() + '/'
+        tpot_obj.periodic_checkpoint_folder = tmpdir
+        # reset _pareto_front to rasie exception
+        tpot_obj._pareto_front = None
 
-        tpot_obj._save_periodic_pipeline()
+        tpot_obj._save_periodic_pipeline(1)
         our_file.seek(0)
 
         assert_in('Failed saving periodic pipeline, exception', our_file.read())
-        # clean up
-        for f in os.listdir('./'):
-            if search('pipeline_', f):
-                os.remove(os.path.join('./', f))
+        #clean up
+        rmtree(tmpdir)
+
 
 def test_save_periodic_pipeline_2():
     """Assert that _save_periodic_pipeline creates the checkpoint folder and exports to it if it didn't exist"""
@@ -1111,20 +1190,49 @@ def test_save_periodic_pipeline_2():
         tpot_obj._last_pipeline_write = datetime.now()
         sleep(0.11)
         tpot_obj._output_best_pipeline_period_seconds = 0.1
-        tpot_obj.periodic_checkpoint_folder = './tmp'
-        tpot_obj._save_periodic_pipeline()
+        tmpdir = mkdtemp() + '_test/'
+        tpot_obj.periodic_checkpoint_folder = tmpdir
+        tpot_obj._save_periodic_pipeline(1)
         our_file.seek(0)
 
         msg = our_file.read()
-        expected_filepath_prefix = os.path.join('./tmp', 'pipeline_')
-        assert_in('Saving best periodic pipeline to ' + expected_filepath_prefix, msg)
-        assert_in('Created new folder to save periodic pipeline: ./tmp', msg)
 
-        # clean up
-        for f in os.listdir('./tmp'):
-            if search('pipeline_', f):
-                os.remove(os.path.join('./tmp', f))
-        os.rmdir('./tmp')
+        assert_in('Saving periodic pipeline from pareto front to {}'.format(tmpdir), msg)
+        assert_in('Created new folder to save periodic pipeline: {}'.format(tmpdir), msg)
+
+        #clean up
+        rmtree(tmpdir)
+
+
+def test_check_periodic_pipeline_3():
+    """Assert that the _save_periodic_pipeline does not export periodic pipeline if the pipeline has been saved before."""
+    tpot_obj = TPOTClassifier(
+        random_state=42,
+        population_size=1,
+        offspring_size=2,
+        generations=1,
+        verbosity=0,
+        config_dict='TPOT light'
+    )
+    tpot_obj.fit(training_features, training_target)
+    with closing(StringIO()) as our_file:
+        tpot_obj._file = our_file
+        tpot_obj.verbosity = 3
+        tpot_obj._exported_pipeline_text = []
+        tpot_obj._last_pipeline_write = datetime.now()
+        sleep(0.11)
+        tpot_obj._output_best_pipeline_period_seconds = 0
+        tmpdir = mkdtemp() + '/'
+        tpot_obj.periodic_checkpoint_folder = tmpdir
+        # export once before
+        tpot_obj._save_periodic_pipeline(1)
+
+        tpot_obj._save_periodic_pipeline(2)
+
+        our_file.seek(0)
+        assert_in('Periodic pipeline was not saved, probably saved before...', our_file.read())
+    #clean up
+    rmtree(tmpdir)
 
 
 def test_fit_predict():
@@ -1321,10 +1429,15 @@ def test_evaluate_individuals():
         config_dict='TPOT light'
     )
     tpot_obj._fit_init()
+    def pareto_eq(ind1, ind2):
+        return np.allclose(ind1.fitness.values, ind2.fitness.values)
+
+    tpot_obj._pareto_front = ParetoFront(similar=pareto_eq)
 
     tpot_obj._pbar = tqdm(total=1, disable=True)
     pop = tpot_obj._toolbox.population(n=10)
-    fitness_scores = tpot_obj._evaluate_individuals(pop, training_features, training_target)
+    pop = tpot_obj._evaluate_individuals(pop, training_features, training_target)
+    fitness_scores = [ind.fitness.values for ind in pop]
 
     for deap_pipeline, fitness_score in zip(pop, fitness_scores):
         operator_count = tpot_obj._operator_count(deap_pipeline)
@@ -1351,10 +1464,15 @@ def test_evaluate_individuals_2():
         config_dict='TPOT light'
     )
     tpot_obj._fit_init()
+    def pareto_eq(ind1, ind2):
+        return np.allclose(ind1.fitness.values, ind2.fitness.values)
+
+    tpot_obj._pareto_front = ParetoFront(similar=pareto_eq)
 
     tpot_obj._pbar = tqdm(total=1, disable=True)
     pop = tpot_obj._toolbox.population(n=10)
-    fitness_scores = tpot_obj._evaluate_individuals(pop, training_features, training_target)
+    pop = tpot_obj._evaluate_individuals(pop, training_features, training_target)
+    fitness_scores = [ind.fitness.values for ind in pop]
 
     for deap_pipeline, fitness_score in zip(pop, fitness_scores):
         operator_count = tpot_obj._operator_count(deap_pipeline)
@@ -1561,6 +1679,7 @@ def test_check_dataset():
         verbosity=0,
         config_dict='TPOT light'
     )
+    tpot_obj._fit_init()
 
     ret_features, ret_target = tpot_obj._check_dataset(training_features, training_target)
     assert np.allclose(ret_features, training_features)
@@ -1577,6 +1696,7 @@ def test_check_dataset_2():
         verbosity=0,
         config_dict='TPOT light'
     )
+    tpot_obj._fit_init()
     test_sample_weight = list(range(1, len(training_target)+1))
     ret_features, ret_target = tpot_obj._check_dataset(training_features, training_target, test_sample_weight)
     test_sample_weight[0] = 'opps'
@@ -1594,6 +1714,7 @@ def test_check_dataset_3():
         verbosity=0,
         config_dict='TPOT light'
     )
+    tpot_obj._fit_init()
     test_sample_weight = list(range(1, len(training_target)+1))
     ret_features, ret_target = tpot_obj._check_dataset(training_features, training_target, test_sample_weight)
     test_sample_weight[0] = np.nan
@@ -1611,6 +1732,7 @@ def test_check_dataset_4():
         verbosity=0,
         config_dict='TPOT light'
     )
+    tpot_obj._fit_init()
     test_sample_weight = list(range(1, len(training_target)))
     assert_raises(ValueError, tpot_obj._check_dataset, training_features, training_target, test_sample_weight)
 
@@ -1625,7 +1747,7 @@ def test_check_dataset_5():
         verbosity=0,
         config_dict='TPOT light'
     )
-
+    tpot_obj._fit_init()
     ret_features = tpot_obj._check_dataset(training_features, target=None)
     assert np.allclose(ret_features, training_features)
 
@@ -1768,6 +1890,29 @@ def test_sparse_matrix_5():
     tpot_obj.fit(sparse_features, sparse_target)
 
 
+def test_source_decode():
+    """Assert that the source_decode can decode operator source and import operator class."""
+    import_str, op_str, op_obj = source_decode("sklearn.linear_model.LogisticRegression")
+    from sklearn.linear_model import LogisticRegression
+    assert import_str == "sklearn.linear_model"
+    assert op_str == "LogisticRegression"
+    assert op_obj == LogisticRegression
+
+
+def test_source_decode_2():
+    """Assert that the source_decode return None when sourcecode is not available."""
+    import_str, op_str, op_obj = source_decode("sklearn.linear_model.LogisticReg")
+    from sklearn.linear_model import LogisticRegression
+    assert import_str == "sklearn.linear_model"
+    assert op_str == "LogisticReg"
+    assert op_obj is None
+
+
+def test_source_decode_3():
+    """Assert that the source_decode raise ImportError when sourcecode is not available and verbose=3."""
+    assert_raises(ImportError, source_decode, "sklearn.linear_model.LogisticReg", 3)
+
+
 def test_tpot_operator_factory_class():
     """Assert that the TPOT operators class factory."""
     test_config_dict = {
@@ -1802,7 +1947,7 @@ def test_tpot_operator_factory_class():
     assert len(tpot_argument_list) == 9
     assert tpot_operator_list[0].root is True
     assert tpot_operator_list[1].root is False
-    assert tpot_operator_list[2].type() == "Classifier or Regressor"
+    assert tpot_operator_list[2].type() == "Classifier"
     assert tpot_argument_list[1].values == [True, False]
 
 
@@ -1810,6 +1955,10 @@ def test_PolynomialFeatures_exception():
     """Assert that TPOT allows only one PolynomialFeatures operator in a pipeline."""
 
     tpot_obj._pbar = tqdm(total=1, disable=True)
+    def pareto_eq(ind1, ind2):
+        return np.allclose(ind1.fitness.values, ind2.fitness.values)
+
+    tpot_obj._pareto_front = ParetoFront(similar=pareto_eq)
     # pipeline with one PolynomialFeatures operator
     pipeline_string_1 = (
         'LogisticRegression(PolynomialFeatures'
@@ -1836,9 +1985,10 @@ def test_PolynomialFeatures_exception():
     for pipeline in pipelines:
         initialize_stats_dict(pipeline)
 
-    fitness_scores = tpot_obj._evaluate_individuals(pipelines, pretest_X, pretest_y)
-    known_scores = [(2, 0.94000000000000006), (5000.0, -float('inf'))]
-    assert np.allclose(known_scores, fitness_scores)
+    pop = tpot_obj._evaluate_individuals(pipelines, pretest_X, pretest_y)
+    fitness_scores = [ind.fitness.values for ind in pop]
+    assert fitness_scores[0][0] == 2
+    assert fitness_scores[1][0] == 5000.0
 
 
 def test_pick_two_individuals_eligible_for_crossover():
@@ -2127,7 +2277,7 @@ def test_varOr_3():
 
 def test_operator_type():
     """Assert that TPOT operators return their type, e.g. 'Classifier', 'Preprocessor'."""
-    assert TPOTSelectPercentile.type() == "Preprocessor or Selector"
+    assert TPOTSelectPercentile.type() == "Selector"
 
 
 def test_gen():
